import multiprocessing
import os
import numpy as np
import pandas as pd


def initialise_multithread(num_cores=-1):
    """
    Initialise pool workers for multi processing
    :param num_cores:
    :return:
    """
    if (num_cores == -1) or (num_cores >= multiprocessing.cpu_count()):
        num_cores = multiprocessing.cpu_count() - 1
    p = multiprocessing.Pool(num_cores)
    return p


def create_directory(directory_path):
    """
    Create a directory if path doesn't exists
    :param directory_path:
    :return:
    """
    if os.path.exists(directory_path):
        return None
    else:
        try:
            os.makedirs(directory_path)
        except:
            # in case another machine created the path meanwhile !:(
            return None
        return directory_path


def save_train_duration(file_name, test_duration):
    """
    Save training time
    :param file_name:
    :param test_duration:
    :return:
    """
    res = pd.DataFrame(data=np.zeros((1, 1), dtype=np.float), index=[0],
                       columns=['train_duration'])
    res['train_duration'] = test_duration
    res.to_csv(file_name, index=False)


def save_test_duration(file_name, test_duration):
    """
    Save test time
    :param file_name:
    :param test_duration:
    :return:
    """
    res = pd.DataFrame(data=np.zeros((1, 1), dtype=np.float), index=[0],
                       columns=['test_duration'])
    res['test_duration'] = test_duration
    res.to_csv(file_name, index=False)



import math
from time import time

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from tqdm import tqdm

from .data_processor import uniform_scaling

name = "RegressorTools"

classical_ml_models = ["xgboost", "svr", "random_forest"]
deep_learning_models = ["fcn", "resnet", "inception"]
tsc_models = ["rocket"]
linear_models = ["lr", "ridge"]
all_models = classical_ml_models + deep_learning_models + tsc_models


def fit_regressor(output_directory, regressor_name, X_train, y_train,
                  X_val=None, y_val=None, itr=1):
    """
    This is a function to fit a regression model given the name and data
    :param output_directory:
    :param regressor_name:
    :param X_train:
    :param y_train:
    :param X_val:
    :param y_val:
    :param itr:
    :return:
    """
    print("[{}] Fitting regressor".format(name))
    start_time = time()

    input_shape = X_train.shape[1:]

    regressor = create_regressor(regressor_name, input_shape, output_directory, itr)
    if (X_val is not None) and (regressor_name in deep_learning_models):
        regressor.fit(X_train, y_train, X_val, y_val)
    else:
        regressor.fit(X_train, y_train)
    elapsed_time = time() - start_time
    print("[{}] Regressor fitted, took {}s".format(name, elapsed_time))
    return regressor


def create_regressor(regressor_name, input_shape, output_directory, verbose=1, itr=1):
    """
    This is a function to create the regression model
    :param regressor_name:
    :param input_shape:
    :param output_directory:
    :param verbose:
    :param itr:
    :return:
    """
    print("[{}] Creating regressor".format(name))
    # SOTA TSC deep learning
    if regressor_name == "resnet":
        from models.deep_learning import resnet
        return resnet.ResNetRegressor(output_directory, input_shape, verbose)
    if regressor_name == "fcn":
        from models.deep_learning import fcn
        return fcn.FCNRegressor(output_directory, input_shape, verbose)
    if regressor_name == "inception":
        from models.deep_learning import inception
        return inception.InceptionTimeRegressor(output_directory, input_shape, verbose)

    if regressor_name == "rocket":
        from models import rocket
        return rocket.RocketRegressor(output_directory)

    # classical ML models
    if regressor_name == "xgboost":
        from models.classical_models import XGBoostRegressor
        kwargs = {"n_estimators": 100,
                  "n_jobs": 0,
                  "learning_rate": 0.1,
                  "random_state": itr - 1,
                  "verbosity  ": verbose}
        return XGBoostRegressor(output_directory, verbose, kwargs)
    if regressor_name == "random_forest":
        from models.classical_models import RFRegressor
        kwargs = {"n_estimators": 100,
                  "n_jobs": -1,
                  "random_state": itr - 1,
                  "verbose": verbose}
        return RFRegressor(output_directory, verbose, kwargs)
    if regressor_name == "svr":
        from models.classical_models import SVRRegressor
        return SVRRegressor(output_directory, verbose)

    # linear models
    if regressor_name == "lr":
        from models.classical_models import LinearRegressor
        kwargs = {"fit_intercept": True,
                  "normalize": False,
                  "n_jobs": -1}
        return LinearRegressor(output_directory, kwargs, type=regressor_name)
    if regressor_name == "ridge":
        from models.classical_models import LinearRegressor
        kwargs = {"fit_intercept": True,
                  "normalize": False}
        return LinearRegressor(output_directory, kwargs, type=regressor_name)


def process_data(X, min_len, scaler=None):
    """
    This is a function to process the data, i.e. convert dataframe to numpy array
    :param X:
    :param min_len:
    :param normalise:
    :return:
    """
    tmp = []
    for i in tqdm(range(len(X))):
        _x = X.iloc[i, :].copy(deep=True)

        # 1. find the maximum length of each dimension
        all_len = [len(y) for y in _x]
        max_len = max(all_len)

        # 2. adjust the length of each dimension
        _y = []
        for y in _x:
            # 2.1 fill missing values
            if y.isnull().any():
                y = y.interpolate(method='linear', limit_direction='both')

            # 2.2. if length of each dimension is different, uniformly scale the shorted one to the max length
            if len(y) < max_len:
                y = uniform_scaling(y, max_len)
            _y.append(y)
        _y = np.array(np.transpose(_y))

        # 3. adjust the length of the series, chop of the longer series
        _y = _y[:min_len, :]

        # 4. normalise the series
        if scaler is not None:
            scaler.fit(_y)
            _y = scaler.transform(_y)

        tmp.append(_y)
    X = np.array(tmp)
    return X, scaler


def calculate_regression_metrics(y_true, y_pred, y_true_val=None, y_pred_val=None):
    """
    This is a function to calculate metrics for regression.
    The metrics being calculated are RMSE and MAE.
    :param y_true:
    :param y_pred:
    :param y_true_val:
    :param y_pred_val:
    :return:
    """
    res = pd.DataFrame(data=np.zeros((1, 2), dtype=np.float), index=[0],
                       columns=['rmse', 'mae'])
    res['rmse'] = math.sqrt(mean_squared_error(y_true, y_pred))
    res['mae'] = mean_absolute_error(y_true, y_pred)

    if not y_true_val is None:
        # this is useful when transfer learning is used with cross validation
        res['rmse_val'] = math.sqrt(mean_squared_error(y_true_val, y_pred_val))
        res['mae_val'] = mean_absolute_error(y_true_val, y_pred_val)

    return res



from time import time

name = "TransformerTools"

transformers = ["pca", "fpca", "fpca_bspline"]


def fit_transformer(transformer_name, X_train, flatten=False, **kwargs):
    """
    Fit a transformer for a set of time series
    :param transformer_name:
    :param X_train:
    :param flatten:
    :param kwargs:
    :return:
    """
    print("[{}] Fitting transformer".format(name))
    start_time = time()

    if flatten:
        # if flatten, do not transform per dimension
        X_train = X_train.reshape(X_train.shape[0], X_train.shape[1] * X_train.shape[2], 1)
    transformer = create_transformer(transformer_name, **kwargs)

    X_train_transformed = transformer.fit_transform(X_train)

    elapsed_time = time() - start_time
    print("[{}] Transformer fitted, took {}s".format(name, elapsed_time))
    return X_train_transformed, transformer


def create_transformer(transformer_name, **kwargs):
    """
    Create a transformer
    :param transformer_name:
    :param kwargs:
    :return:
    """
    print("[{}] Creating transformer".format(name))
    if transformer_name == "pca":
        from transform.transformers import PCATransformer
        return PCATransformer(**kwargs)
    if transformer_name == "fpca":
        from transform.transformers import FPCATransformer
        return FPCATransformer(**kwargs)
    if transformer_name == "fpca_bspline":
        from transform.transformers import FPCATransformer
        return FPCATransformer(**kwargs)

    from transform.transformers import TimeSeriesTransformer
    return TimeSeriesTransformer(**kwargs)



from transformers import DataCollatorForLanguageModeling, DataCollatorWithPadding, AutoTokenizer, AutoModelForCausalLM, BitsAndBytesConfig, Trainer, TrainingArguments
from accelerate import FullyShardedDataParallelPlugin, Accelerator
import torch
from torch.distributed.fsdp.fully_sharded_data_parallel import FullOptimStateDictConfig, FullStateDictConfig
from peft import LoraConfig, PeftModel, prepare_model_for_kbit_training, get_peft_model, TaskType
from sklearn.utils.class_weight import compute_class_weight
from sklearn import preprocessing
import numpy as np
from .abba import XABBA
import pickle
import os
import torch
from datasets.dataset_dict import DatasetDict
from peft import PeftModel
from .utils.fundamentals import *
from .utils.data_loader import load_from_tsfile_to_dataframe
from .utils.regressor_tools import process_data, fit_regressor, calculate_regression_metrics
from .utils.tools import create_directory
from .utils.transformer_tools import fit_transformer
from sklearn.model_selection import train_test_split
from datetime import datetime
from datasets.dataset_dict import DatasetDict
from datasets import Dataset

def save_abba(model, filename):
    pickle.dump(model, file = open(filename, "wb"))

def load_abba(filename):
    return pickle.load(open(filename, "rb"))


# Python program to convert a list to string
def ListToString(s):
    # initialize an empty string
    str1 = ""
    # traverse in the string
    for ele in s:
        str1 += ele + ' '
    # return string
    return str1

# Python program to convert a string to list
def StringToList(s):
    # initialize an empty string
    str1 = []
    # traverse in the string
    for ele in range(int(len(s)/2)):
        str1.append(s[ele*2])
    # return string
    return str1


class LLMABBA:
    def __init__(self,
                 abba_tol = 0.05,
                 abba_init = 'agg',
                 abba_alpha = 0.01,
                 bits_for_len = 16,
                 bits_for_inc = 16,
                 abba_scl = 3,
                 abba_verbose = 0,
                 lora_r = 16,
                 lora_alpha = 16,
                 target_modules = None,
                 modules_to_save = ["embed_tokens"],
                 lora_dropout = 0.05,
                 quant_process = True
                 ):

        self.abba_tol = abba_tol
        self.abba_init = abba_init
        self.abba_alpha = abba_alpha
        self.bits_for_len = bits_for_len
        self.bits_for_inc = bits_for_inc
        self.abba_scl = abba_scl
        self.abba_verbose = abba_verbose
        self.lora_r = lora_r 
        self.lora_alpha = lora_alpha
        self.quant_process = quant_process
        if target_modules is not None:
            self.target_modules = target_modules
        else:
            self.target_modules = [
                            "q_proj",
                            "k_proj",
                            "v_proj",
                            "o_proj",
                            "gate_proj",
                            "up_proj",
                            "down_proj",
                            "lm_head",
                        ]

        self.modules_to_save = modules_to_save
        self.lora_dropout = lora_dropout


    def tokenize(self, prompt_input):
        result = self.model_tokenizer(
            prompt_input,
            truncation=True,
            max_length=self.max_len,
            padding="max_length",
        )
        result["labels"] = result["input_ids"].copy()
        return result

    def build(self, data_point):
        """Build the Prompt Data"""
        full_prompt = f"""{self.prompt}
        ### Symbolic Series: {data_point["text_inputs"]}
        ### Results: {data_point["text_outputs"]}
        """
        return self.tokenize(full_prompt)


    def process(self, project_name, data, task, prompt, model_tokenizer,
                seq_len_pre = 24, scalar="z-score",
                seq_len_post = 24, alphabet_set=None)-> None:
        """Load data and process data"""

        if alphabet_set is None:
            vocab_list = model_tokenizer.get_vocab()
            alphabet_set = list(vocab_list.keys())
            print("Using LLM tokens as the alphabet_set")
        else:
            print("Using self-defined alphabet_set")

        self.project_name = project_name
        self.task = task
        self.alphabet_set = alphabet_set
        self.prompt = prompt
        self.model_tokenizer = model_tokenizer
        self.seq_len_pre = seq_len_pre
        self.seq_len_post = seq_len_post

        # Initialize the StandardScaler
        if scalar == "min-max":
            self.scaler = preprocessing.MinMaxScaler(feature_range=(0, 1))
        elif scalar == "z-score":
            self.scaler =  preprocessing.StandardScaler()

        # Initialize the ABBA block
        self.xabba = XABBA(tol=self.abba_tol, init=self.abba_init,
                           alpha=self.abba_alpha, scl=self.abba_scl, verbose=0)

        if self.task == "classification":
            # Fit the scaler to the training data and transform it
            X_data = self.scaler.fit_transform(data['X_data'])
            Y_data = data['Y_data']

            symbols = self.xabba.fit_transform(X_data, alphabet_set=self.alphabet_set)
            reconstruction = self.xabba.inverse_transform(symbols)
            # reconst_same_shape = self.xabba.recast_shape(reconstruction)  # recast into original shape

            symbols_convert = []
            for i_data in range(len(symbols)):
                symbols_convert.append(listToString(list(symbols[i_data])))

            train_data_symbolic, val_data_symbolic, train_target_symbolic, val_target_symbolic = train_test_split(
                symbols_convert, Y_data, test_size=0.2)


        elif self.task == "regression":

            # Fit the scaler to the training data and transform it
            min_len = np.inf
            for i in range(len(data['X_data'])):
                x = data['X_data'].iloc[i, :]
                all_len = [len(y) for y in x]
                min_len = min(min(all_len), min_len)
            # print("[{}] Minimum length: {}".format(module, min_len))

            X_data, self.scaler = process_data(data['X_data'], scaler=self.scaler, min_len=min_len)
            Y_data = data['Y_data']

            symbols = self.xabba.fit_transform(X_data, alphabet_set=self.alphabet_set)
            # reconstruction = self.xabba.inverse_transform(symbols)
            # reconst_same_shape = self.xabba.recast_shape(reconstruction)  # recast into original shape

            symbols_convert = []
            for i_data in range(len(symbols)):
                symbols_convert.append(listToString(list(symbols[i_data])))

            train_data_symbolic, val_data_symbolic, train_target_symbolic, val_target_symbolic = train_test_split(
                symbols_convert, Y_data, test_size=0.2)

        elif self.task == "forecasting":
            # Fit the scaler to the training data and transform it
            self.scaler.fit(data)
            data_scaled = self.scaler.transform(data)

            X_Train_data_patch = np.zeros(
                [data_scaled.shape[0] - (self.seq_len_pre + self.seq_len_post), self.seq_len_pre,
                 data_scaled.shape[1]],
                dtype=float)
            Y_Train_data_patch = np.zeros(
                [data_scaled.shape[0] - (self.seq_len_pre + self.seq_len_post), self.seq_len_post,
                 data_scaled.shape[1]],
                dtype=float)

            for i_data_patch in range(data_scaled.shape[0] - (self.seq_len_pre + self.seq_len_post)):
                X_Train_data_patch[i_data_patch, :, :] = data_scaled[i_data_patch:i_data_patch + self.seq_len_pre,
                                                         :]
                Y_Train_data_patch[i_data_patch, :, :] = data_scaled[
                                                         i_data_patch + self.seq_len_pre:i_data_patch + self.seq_len_pre + self.seq_len_post,
                                                         :]

            symbols_train_data = self.xabba.fit_transform(X_Train_data_patch, alphabet_set=self.alphabet_set, llm_split='Pre')
            # reconstruction_train_data = qabba.inverse_transform(symbols_train_data)
            # train_data_same_shape = qabba.recast_shape(reconstruction_train_data,
            #                                            llm_split='Pre')  # recast into original shape

            symbols_train_target, params_train_target = self.xabba.transform(Y_Train_data_patch, llm_split='Post')
            # reconstruction_train_target = qabba.inverse_transform(symbols_train_target, params_train_target)
            # train_target_same_shape = qabba.recast_shape(reconstruction_train_target, recap_shape=Y_Train_data_patch.shape,
            #                                              llm_split='Post')  # recast into original shape

            train_data_symbolic = []
            for i_data in range(len(symbols_train_data)):
                train_data_symbolic.append(listToString(list(symbols_train_data[i_data])))

            train_target_symbolic = []
            for i_data in range(len(symbols_train_target)):
                train_target_symbolic.append(listToString(list(symbols_train_target[i_data])))

            arranged_seq = np.random.randint(len(train_data_symbolic), size=int(len(train_data_symbolic) * 0.2))

            val_data_symbolic = [train_data_symbolic[index] for index in arranged_seq]
            val_target_symbolic = [train_target_symbolic[index] for index in arranged_seq]

        else:
            raise NotImplementedError("Method is not implemented, please contact the maintenance team.")

        data_TS = DatasetDict({
            'train': Dataset.from_dict({'text_outputs': train_target_symbolic, 'text_inputs': train_data_symbolic}),
            'val': Dataset.from_dict({'text_outputs': val_target_symbolic, 'text_inputs': val_data_symbolic})
        })

        # Check whether the specified path exists or not
        path = "../save/" + self.project_name + "/"
        isExist = os.path.exists(path)
        if not isExist:
            # Create a new directory because it does not exist
            os.makedirs(path)
            print("The new project directory is created!")

        #### Saving Scaler and ABBA
        output_scaler = open(str("../save/" + self.project_name + "/" + self.task  + "_Scaler" + "_save.pkl"), 'wb')

        str1 = pickle.dumps(self.scaler)
        output_scaler.write(str1)
        output_scaler.close()

        # curr_loc = os.path.dirname(os.path.realpath(__file__))
        save_abba(self.xabba, str("../save/" + self.project_name + "/" + self.task  + "_ABBA" + "_save.pkl"))

        tokenized_train_dataset = data_TS['train'].map(self.build)
        tokenized_val_dataset = data_TS['val'].map(self.build)

        tokenized_train_dataset.set_format("torch")
        tokenized_val_dataset.set_format("torch")

        return tokenized_train_dataset, tokenized_val_dataset



    def model(self, peft_file = None, model_name = None, max_len = 2048) -> None:

        self.model_name = model_name
        self.max_len = max_len

        if peft_file is None:
            fsdp_plugin = FullyShardedDataParallelPlugin(
                state_dict_config=FullStateDictConfig(offload_to_cpu=True, rank0_only=False),
                optim_state_dict_config=FullOptimStateDictConfig(offload_to_cpu=True, rank0_only=False),
            )
            accelerator = Accelerator(fsdp_plugin=fsdp_plugin)

            """Model selection with parameters"""
            if self.quant_process is True:
                ##  Quantization Coonfig
                quantization_config = BitsAndBytesConfig(
                    # Load the model with 4-bit quantization
                    load_in_4bit=True,
                    # Use double quantization
                    bnb_4bit_use_double_quant=True,
                    # Use 4-bit Normal Float for storing the base model weights in GPU memory
                    bnb_4bit_quant_type="nf4",
                    # De-quantize the weights to 16-bit (Brain) float before the forward/backward pass
                    bnb_4bit_compute_dtype=torch.bfloat16,
                )

                model_tokenizer = AutoTokenizer.from_pretrained(
                    model_name,
                    model_max_len=max_len,
                    padding_side="right",
                    truncation=True,
                    add_eos_token=True,
                )

                model_tokenizer.padding_side = 'right'
                model_tokenizer.pad_token = model_tokenizer.eos_token
                model_tokenizer.pad_token_id = model_tokenizer.eos_token_id

                model_input = AutoModelForCausalLM.from_pretrained(
                    pretrained_model_name_or_path=model_name,
                    quantization_config=quantization_config,
                    # device_map="auto",
                    trust_remote_code=True,
                    torch_dtype=torch.bfloat16,
                )

                model_input.config.pad_token_id = model_input.config.eos_token_id
                model_input = prepare_model_for_kbit_training(model_input)

            else:

                model_tokenizer = AutoTokenizer.from_pretrained(
                    model_name,
                    model_max_len=max_len,
                    padding_side="right",
                    truncation=True,
                    add_eos_token=True,
                )

                model_tokenizer.padding_side = 'right'
                model_tokenizer.pad_token = model_tokenizer.eos_token
                model_tokenizer.pad_token_id = model_tokenizer.eos_token_id
                model_input = AutoModelForCausalLM.from_pretrained(
                    pretrained_model_name_or_path=model_name,
                    # device_map="auto",
                    trust_remote_code=True,
                )
                # Data collator for padding a batch of examples to the maximum length seen in the batch
                self.model_data_collator = DataCollatorWithPadding(tokenizer=model_tokenizer)

                model_input.config.pad_token_id = model_input.config.eos_token_id
                model_input = prepare_model_for_kbit_training(model_input)

            #####   Loading LoRA config and PEFT model
            lora_config = LoraConfig(
                task_type=TaskType.SEQ_CLS,
                r=self.lora_r,  # the dimension of the low-rank matrices
                lora_alpha=self.lora_alpha,  # scaling factor for LoRA activations vs pre-trained weight activations
                target_modules=self.target_modules,
                lora_dropout=self.lora_dropout,  # dropout probability of the LoRA layers
                bias='none',  # wether to train bias weights, set to 'none' for attention layers
                modules_to_save=self.modules_to_save,
            )

            model_input = get_peft_model(model_input, lora_config)
            # Apply the accelerator. You can comment this out to remove the accelerator.
            model_input = accelerator.prepare_model(model_input)

            model_input.print_trainable_parameters()
            model_input = model_input.cuda()

        else:
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=torch.bfloat16
            )

            model = AutoModelForCausalLM.from_pretrained(
                model_name,  # Mistral, same as before
                quantization_config=bnb_config,  # Same quantization config as before
                # device_map="auto",
                trust_remote_code=True,
            )

            model_tokenizer = AutoTokenizer.from_pretrained(
                model_name,
                model_max_length=self.max_len,
                padding_side="right",
                truncation=True,
                add_eos_token=True,
            )

            model_tokenizer.padding_side = 'right'
            model_tokenizer.pad_token = model_tokenizer.eos_token
            model_tokenizer.pad_token_id = model_tokenizer.eos_token_id
            print(len(model_tokenizer))

            model_input = PeftModel.from_pretrained(model, peft_file)

        self.model_tokenizer = model_tokenizer
        return model_input, model_tokenizer



    def train(self, model_input, num_epochs, output_dir, train_dataset, val_dataset):
        """Train with validation"""
        if torch.cuda.device_count() > 1:  # If more than 1 GPU
            model.is_parallelizable = True
            model.model_parallel = True


        run_name = self.task + "-" + self.model_name

        trainer = Trainer(
            model=model_input,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            args=TrainingArguments(
                output_dir=output_dir,
                warmup_steps=5,
                per_device_train_batch_size=4,
                gradient_checkpointing=True,
                gradient_accumulation_steps=4,
                learning_rate=2e-4,  # Want about 10x smaller than the Mistral learning rate
                bf16=False,
                optim="paged_adamw_8bit",
                num_train_epochs=num_epochs,
                weight_decay=0.00005,
                evaluation_strategy="epoch",
                save_strategy="epoch",
                logging_strategy="epoch",
                logging_dir="./logs",
                load_best_model_at_end=True,
                report_to="wandb",
                save_total_limit=1,
                do_eval=True,  # Perform evaluation at the end of training
                # report_to="wandb",           # Comment this out if you don't want to use weights & baises
                run_name=f"{run_name}-{datetime.now().strftime('%Y-%m-%d-%H-%M')}"  # Name of the W&B run (optional)
            ),
            data_collator=DataCollatorForLanguageModeling(self.model_tokenizer, mlm=False),
        )

        model_input.config.use_cache = False  # silence the warnings. Please re-enable for inference!
        trainer.train()
        return self


    def inference(self, project_name, data, task, prompt, ft_model, model_tokenizer, scalar="z-score",
                  seq_len_pre = 24, seq_len_post = 24,
                  llm_max_length=1024, llm_repetition_penalty=1.7,
                  llm_temperature=0.0, llm_max_new_tokens=256,
                )-> None:
        """Inference """

        # self.task = task
        # self.model_tokenizer = model_tokenizer


        # Initialize the StandardScaler
        if scalar == "min-max":
            self.scaler = preprocessing.MinMaxScaler(feature_range=(0, 1))
        elif scalar == "z-score":
            self.scaler =  preprocessing.StandardScaler()

        # Initialize the ABBA block
        self.xabba = XABBA(tol=self.abba_tol, init=self.abba_init,
                           alpha=self.abba_alpha, scl=self.abba_scl, verbose=0)

        with open("../save/" + project_name + "/" + task  + "_ABBA" + "_save.pkl", 'rb') as file:
            self.xabba = pickle.loads(file.read())

        with open(str("../save/" + project_name + "/" + task  + "_Scaler" + "_save.pkl"), 'rb') as file:
            self.scaler = pickle.loads(file.read())

        if (task == "classification") or (task == "regression"):
            data_scaled = self.scaler.transform(data)
            symbols_test_data, params_test_data = self.xabba.transform(data_scaled)

        elif task == "forecasting":
            data_scaled = self.scaler.transform(data)
            symbols_test_data, params_test_data = self.xabba.transform(data_scaled, llm_split='Post')
            # reconstruction_test_data = qabba_load.inverse_transform(symbols_test_data, params_test_data)
            # test_data_same_shape = qabba_load.recast_shape(reconstruction_test_data,
            #                                                recap_shape=scaled_test_data.shape,
            #                                                llm_split='Post')  # recast into original shape

        else:
            print("The task is UNKNOWN!")
            pass


        print(symbols_test_data[0])
        test_data_symbolic = ListToString(list(symbols_test_data[0]))

        if prompt is not None:
            test_prompt = f"""{prompt}
            ### Symbolic Series: {test_data_symbolic}
            ### Results:
            """
        else:
            print("Please provide your Prompt!")

        model_input_text = model_tokenizer(test_prompt, return_tensors="pt").to('cuda')

        llm_out = model_tokenizer.decode(
            ft_model.generate(
                **model_input_text,
                max_new_tokens=llm_max_new_tokens,
                max_length=llm_max_length,
                repetition_penalty=llm_repetition_penalty,
                temperature=llm_temperature,
            )[0],
            skip_special_tokens=True
        )


        if (task == "classification") or (task == "regression"):
            return llm_out

        elif task == "forecasting":
            split_content = llm_out.split('Results:')
            listed_split_content = StringToList(split_content[1])

            reconst_test_output = self.xabba.inverse_transform(listed_split_content, params_test_data)  # convert into array
            reconst_same_shape_output = self.xabba.recast_shape(reconst_test_output, recap_shape=[seq_len_post, 7], llm_split='Post')
            forecasting_out = self.scaler.inverse_transform(reconst_same_shape_output)

            return forecasting_out




# Python program to convert a list to string
def listToString(s):
    # initialize an empty string
    str1 = ""
    # traverse in the string
    for ele in s:
        str1 += ele + ' '
    # return string
    return str1

# Python program to convert a list to string
def listToString_blank(s):
    # initialize an empty string
    str1 = ""
    # traverse in the string
    for ele in s:
        str1 += ele + ''
    # return string
    return str1

# Python program to convert a list to string
def stringToList(s):
    # initialize an empty string
    str1 = []
    # traverse in the string
    for ele in range(int(len(s)/2)):
        str1.append(s[ele*2])
    # return string
    return str1

def mean(arr):
    return sum(arr) / len(arr)

def cross_correlation(x, y):
    # Calculate means
    x_mean = mean(x)
    y_mean = mean(y)

    # Calculate numerator
    numerator = sum((a - x_mean) * (b - y_mean) for a, b in zip(x, y))

    # Calculate denominators
    x_sq_diff = sum((a - x_mean) ** 2 for a in x)
    y_sq_diff = sum((b - y_mean) ** 2 for b in y)
    denominator = math.sqrt(x_sq_diff * y_sq_diff)
    correlation = numerator / denominator

    return correlation


def find_keys_by_value(dictionary, value):
    return [key for key, val in dictionary.items() if val == value]


def find_values_by_key(dictionary, keys):
    return [dictionary[key] for key in keys]


# def model_preprocessing_function(examples):
#     return model_tokenizer(examples['text'], truncation=True, padding='max_length', max_length=MAX_LENGTH)


def fnirs_process(current_dir, input_folder):

    # column_name = ['AB_I_O', 'AB_PHI_O', 'AB_I_DO', 'AB_PHI_DO', 'CD_I_O', 'CD_PHI_O', 'CD_I_DO', 'CD_PHI_DO', 'label']
    # data_df = pd.DataFrame([], columns=column_name)

    data_all_transformed, target_all_transformed = [], []

    subject_seg = 426
    seg_length = 210 # sampling rate is 5.2Hz, one trial is 2 second length,
    for file_i in range(len(input_folder)): #2):#
        directory = input_folder[file_i]  ## 57: Haptics  41: FordA  33: Earthquakes  24: DistalPhalanxTW
        input_data = current_dir + "/" + directory

        ##   data reading
        data_subject = pd.read_csv(input_data, header=None).to_numpy()
        data_subject = data_subject[1:, :]
        data_subject_size = data_subject.shape

        subject_seg_temp = 0
        seg_length_temp = 0
        AA = np.zeros([subject_seg - 6, data_subject_size[1] - 1])
        BB = np.zeros([seg_length, data_subject_size[1] - 1])
        data_subject_transformed = np.zeros(
            [int(data_subject_size[0] / subject_seg) * int(subject_seg / seg_length), data_subject_size[1] - 1, seg_length])
        target_subject_transformed = np.zeros([int(data_subject_size[0] / subject_seg) * int(subject_seg / seg_length)],
                                              dtype=int)
        for i_subject_seg in range(int(data_subject_size[0] / subject_seg)):
            AA = data_subject[i_subject_seg * subject_seg + 3:(i_subject_seg + 1) * subject_seg - 3, :]
            subject_seg_temp += 1
            for i_seg_length in range(int(subject_seg / seg_length)):
                BB = AA[i_seg_length * seg_length:(i_seg_length + 1) * seg_length, :-1]
                data_subject_transformed[seg_length_temp, :, :] = np.transpose(BB)
                target_subject_transformed[seg_length_temp] = AA[i_seg_length * seg_length, -1]
                seg_length_temp += 1

        data_all_transformed.extend(data_subject_transformed)
        target_all_transformed.extend(target_subject_transformed)

    return np.array(data_all_transformed), np.array(target_all_transformed)


def fnirs_process_slide_window(current_dir, input_folder):

    # column_name = ['AB_I_O', 'AB_PHI_O', 'AB_I_DO', 'AB_PHI_DO', 'CD_I_O', 'CD_PHI_O', 'CD_I_DO', 'CD_PHI_DO', 'label']
    # data_df = pd.DataFrame([], columns=column_name)

    data_all_transformed, target_all_transformed = [], []

    subject_seg = 150  # sampling rate is 5.2Hz, we set 30 second length,
    for file_i in range(len(input_folder)): #2):#
        directory = input_folder[file_i]
        input_data = current_dir + "/" + directory

        ##   data reading
        data_subject = pd.read_csv(input_data, header=None).to_numpy()
        data_subject = data_subject[1:, :]

        data_subject_size = data_subject.shape

        subject_seg_temp = 0
        AA = np.zeros([subject_seg, data_subject_size[1]])
        data_subject_transformed = np.zeros([int(data_subject_size[0] / subject_seg), subject_seg, data_subject_size[1] - 2])
        target_subject_transformed = np.zeros([int(data_subject_size[0] / subject_seg)],
                                              dtype=int)
        for i_subject_seg in range(int(data_subject_size[0] / subject_seg)):
            AA = data_subject[i_subject_seg * subject_seg:(i_subject_seg + 1) * subject_seg, :]
            data_subject_transformed[subject_seg_temp, :, :] = AA[:, :-2]   #  np.transpose(BB)
            target_subject_transformed[subject_seg_temp] = AA[75, -1]
            subject_seg_temp += 1

        baseline_index = np.where(target_subject_transformed == 0)
        wm_load3_index = np.where(target_subject_transformed == 3)

        baseline_data_subject_transformed = np.squeeze(data_subject_transformed[baseline_index, :, :])
        wm_load3_data_subject_transformed = np.squeeze(data_subject_transformed[wm_load3_index, :, :])

        baseline_target_subject_transformed = target_subject_transformed[baseline_index]
        wm_load3_target_subject_transformed = np.ones_like(target_subject_transformed[wm_load3_index])

        BinaryClass_data_subject_transformed = np.concatenate((baseline_data_subject_transformed, wm_load3_data_subject_transformed), axis = 0)
        BinaryClass_dtarget_subject_transformed = np.concatenate((baseline_target_subject_transformed, wm_load3_target_subject_transformed), axis = 0)

        data_all_transformed.extend(BinaryClass_data_subject_transformed)
        target_all_transformed.extend(BinaryClass_dtarget_subject_transformed)

    return np.array(data_all_transformed), np.array(target_all_transformed)



import numpy as np
import warnings 

def rd_func(x):
    return np.round(x, decimals=0)
    


def clip_func(x, a_min, a_max):
    return np.clip(x, a_min=a_min, a_max=a_max)


    
class QUANT():
    """
    Parameters
    ----------
    bits : int, default=8
        The bitwidth of integer format, the larger it is, the wider range the quantized value can be.
        
    sign : bool, default=1
        Whether or not to quantize the value to symmetric integer range.
    
    zpoint : bool, default=1
        Whether or not to compute the zero point. If `zpoint=0`, then the quantized range must be symmetric.
        
    rd_func : function, default=None
        The rounding function used for the quantization. The default is round to nearest.
        
    clip_range : list, default=None
        The clipping function for the quantization.
        
    epsilon : double, default=1e-12
        When the x is comprised of single value, then the scaling factor will be (b - a + epsilon) / (alpha - beta)
        for mapping [a, b] to [alpha, beta].
        
    Methods
    ----------
    quant(x):
        Method that quantize ``x`` to the user-specific arithmetic format.
        
    """
    def __init__(self, bits=8, sign=1, zpoint=1, rd_func=None, clip_range=None, epsilon=1e-12):
        self.bits = bits
        self.sign = sign
        self.zpoint = zpoint
        
        self.clip_range = clip_range
        self.epsilon = epsilon 
        
        if bits in {8, 16, 32, 64}:
            if bits == 8:
                self.intType = np.int8
                
            elif bits == 16:
                self.intType = np.int16
                
            elif bits == 32:
                self.intType = np.int32
                
            elif bits == 64:
                self.intType = np.int64
                
        else:
            warnings.warn("Current int type not support this bitwidth, use int64 to simulate.")
            self.intType = np.int64
        
        if self.sign == 1:
            if self.zpoint == 1:
                self.alpha_q = -2**(self.bits - 1)
                self.beta_q = 2**(self.bits - 1) - 1
            else:
                self.beta_q = 2**(self.bits - 1) - 1
                self.alpha_q = -self.beta_q
                
        else:
            if self.zpoint == 0:
                self.beta_q = 2**(self.bits - 1) - 1
                self.alpha_q = -self.beta_q
            else:
                raise ValueError('Please set `zpoint` to 0.')
        
            
            
    def __call__(self, x):
        x_min = np.min(x)
        x_max = np.max(x)
        
        if self.sign != 1:
            abs_max = max(abs(x_min), abs(x_max))
            x_min = -abs_max
            x_max = abs_max
            
        self.scaling, self.zpoint = self.compute_scaling(x_min, x_max, self.alpha_q, self.beta_q)
        
        try:
            x_q = self.quantization(x, self.scaling, self.zpoint)
        except:
            self.scaling, self.zpoint = self.compute_scaling(x_min, x_max+self.epsilon, self.alpha_q, self.beta_q)
            x_q = self.quantization(x, self.scaling, self.zpoint)
            
        return x_q
        
    def dequant(self, x_q):
        return self.dequantization(x_q, self.scaling, self.zpoint)
        
    def quantization(self, x, s, z):
        x_q = rd_func((x - z)/s)
            
        if self.clip_range is None:
            x_q = self.intType(clip_func(x_q, a_min=self.alpha_q, a_max=self.beta_q))
        else:
            x_q = self.intType(clip_func(x_q, a_min=self.clip_range[0], a_max=self.beta_q[self.clip_range[1]]))
        
        return x_q

    def dequantization(self, x_q, s, z):
        x_q = x_q.astype(np.float32)
        x = s * x_q + z
        return x

    def compute_scaling(self, alpha, beta, alpha_q, beta_q):
        s = (beta - alpha) / (beta_q - alpha_q)
        z = (alpha * beta_q - beta * alpha_q) / (beta_q - alpha_q)

        return s, z
    


cimport cython
import numpy as np
cimport numpy as np
np.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.binding(True)

cpdef compress(double[:] ts, double tol=0.5, double max_len=np.finfo(float).max):
    """
    Approximate a time series using a continuous piecewise linear function.

    Parameters
    ----------
    ts - numpy ndarray
        Time series as input of numpy array

    Returns
    -------
    pieces - numpy array
        Numpy ndarray with three columns, each row contains length, increment, error for the segment.
    """

    cdef int start = 0
    cdef int end = 1
    cdef int i, len_t = len(ts)
    cdef list pieces = list() # np.empty([0, 3])
    cdef double[:] x = np.arange(0, len_t, dtype=float)
    cdef double epsilon =  np.finfo(float).eps
    
    cdef double t_st, t_ed
    cdef double inc, lastinc, err, lasterr
    
    while end < len_t:
        t_st = ts[start]
        t_ed = ts[end]
        inc = t_ed - t_st
        err = 0.0
        
        err = sum([(t_st + (inc/(end-start))*x[0:end-start+1][i] - ts[start:end+1][i])**2 for i in range(end-start+1)])
        
        if (err <= (tol**2)*(end-start-1) + epsilon) and (end-start-1 < max_len):
            (last_t_ed, lasterr) = (t_ed, err) 
            end = end + 1
        else:
            pieces.append([end-start-1, last_t_ed, lasterr])
            start = end - 1

    pieces.append([end-start-1, last_t_ed, lasterr])

    return np.asarray(pieces)


import numpy as np
import pandas as pd
from tqdm import tqdm

name = "DataLoader"

regression_datasets = ["AustraliaRainfall",
                       "HouseholdPowerConsumption1",
                       "HouseholdPowerConsumption2",
                       "BeijingPM25Quality",
                       "BeijingPM10Quality",
                       "Covid3Month",
                       "LiveFuelMoistureContent",
                       "FloodModeling1",
                       "FloodModeling2",
                       "FloodModeling3",
                       "AppliancesEnergy",
                       "BenzeneConcentration",
                       "NewsHeadlineSentiment",
                       "NewsTitleSentiment",
                       "BIDMC32RR",
                       "BIDMC32HR",
                       "BIDMC32SpO2",
                       "IEEEPPG",
                       "PPGDalia"]


# The following code is adapted from the python package sktime to read .ts file.
class TsFileParseException(Exception):
    """
    Should be raised when parsing a .ts file and the format is incorrect.
    """
    pass


def load_from_tsfile_to_dataframe(full_file_path_and_name, return_separate_X_and_y=True,
                                  replace_missing_vals_with='NaN'):
    """Loads data from a .ts file into a Pandas DataFrame.

    Parameters
    ----------
    full_file_path_and_name: str
        The full pathname of the .ts file to read.
    return_separate_X_and_y: bool
        true if X and Y values should be returned as separate Data Frames (X) and a numpy array (y), false otherwise.
        This is only relevant for data that
    replace_missing_vals_with: str
       The value that missing values in the text file should be replaced with prior to parsing.

    Returns
    -------
    DataFrame, ndarray
        If return_separate_X_and_y then a tuple containing a DataFrame and a numpy array containing the relevant time-series and corresponding class values.
    DataFrame
        If not return_separate_X_and_y then a single DataFrame containing all time-series and (if relevant) a column "class_vals" the associated class values.
    """

    # Initialize flags and variables used when parsing the file
    metadata_started = False
    data_started = False

    has_problem_name_tag = False
    has_timestamps_tag = False
    has_univariate_tag = False
    has_class_labels_tag = False
    has_target_labels_tag = False
    has_data_tag = False

    previous_timestamp_was_float = None
    previous_timestamp_was_int = None
    previous_timestamp_was_timestamp = None
    num_dimensions = None
    is_first_case = True
    instance_list = []
    class_val_list = []
    line_num = 0

    # Parse the file
    # print(full_file_path_and_name)
    with open(full_file_path_and_name, 'r', encoding='utf-8', errors='ignore') as file:
        for line in tqdm(file):
            # print(".", end='')
            # Strip white space from start/end of line and change to lowercase for use below
            line = line.strip().lower()
            # Empty lines are valid at any point in a file
            if line:
                # Check if this line contains metadata
                # Please note that even though metadata is stored in this function it is not currently published externally
                if line.startswith("@problemname"):
                    # Check that the data has not started
                    if data_started:
                        raise TsFileParseException("metadata must come before data")
                    # Check that the associated value is valid
                    tokens = line.split(' ')
                    token_len = len(tokens)

                    if token_len == 1:
                        raise TsFileParseException("problemname tag requires an associated value")

                    problem_name = line[len("@problemname") + 1:]
                    has_problem_name_tag = True
                    metadata_started = True
                elif line.startswith("@timestamps"):
                    # Check that the data has not started
                    if data_started:
                        raise TsFileParseException("metadata must come before data")

                    # Check that the associated value is valid
                    tokens = line.split(' ')
                    token_len = len(tokens)

                    if token_len != 2:
                        raise TsFileParseException("timestamps tag requires an associated Boolean value")
                    elif tokens[1] == "true":
                        timestamps = True
                    elif tokens[1] == "false":
                        timestamps = False
                    else:
                        raise TsFileParseException("invalid timestamps value")
                    has_timestamps_tag = True
                    metadata_started = True
                elif line.startswith("@univariate"):
                    # Check that the data has not started
                    if data_started:
                        raise TsFileParseException("metadata must come before data")

                    # Check that the associated value is valid
                    tokens = line.split(' ')
                    token_len = len(tokens)
                    if token_len != 2:
                        raise TsFileParseException("univariate tag requires an associated Boolean value")
                    elif tokens[1] == "true":
                        univariate = True
                    elif tokens[1] == "false":
                        univariate = False
                    else:
                        raise TsFileParseException("invalid univariate value")

                    has_univariate_tag = True
                    metadata_started = True
                elif line.startswith("@classlabel"):
                    # Check that the data has not started
                    if data_started:
                        raise TsFileParseException("metadata must come before data")

                    # Check that the associated value is valid
                    tokens = line.split(' ')
                    token_len = len(tokens)

                    if token_len == 1:
                        raise TsFileParseException("classlabel tag requires an associated Boolean value")

                    if tokens[1] == "true":
                        class_labels = True
                    elif tokens[1] == "false":
                        class_labels = False
                    else:
                        raise TsFileParseException("invalid classLabel value")

                    # Check if we have any associated class values
                    if token_len == 2 and class_labels:
                        raise TsFileParseException("if the classlabel tag is true then class values must be supplied")

                    has_class_labels_tag = True
                    class_label_list = [token.strip() for token in tokens[2:]]
                    metadata_started = True
                elif line.startswith("@targetlabel"):
                    # Check that the data has not started
                    if data_started:
                        raise TsFileParseException("metadata must come before data")

                    # Check that the associated value is valid
                    tokens = line.split(' ')
                    token_len = len(tokens)

                    if token_len == 1:
                        raise TsFileParseException("targetlabel tag requires an associated Boolean value")

                    if tokens[1] == "true":
                        target_labels = True
                    elif tokens[1] == "false":
                        target_labels = False
                    else:
                        raise TsFileParseException("invalid targetLabel value")

                    has_target_labels_tag = True
                    class_val_list = []
                    metadata_started = True
                # Check if this line contains the start of data
                elif line.startswith("@data"):
                    if line != "@data":
                        raise TsFileParseException("data tag should not have an associated value")

                    if data_started and not metadata_started:
                        raise TsFileParseException("metadata must come before data")
                    else:
                        has_data_tag = True
                        data_started = True
                # If the 'data tag has been found then metadata has been parsed and data can be loaded
                elif data_started:
                    # Check that a full set of metadata has been provided
                    incomplete_regression_meta_data = not has_problem_name_tag or not has_timestamps_tag or not has_univariate_tag or not has_target_labels_tag or not has_data_tag
                    incomplete_classification_meta_data = not has_problem_name_tag or not has_timestamps_tag or not has_univariate_tag or not has_class_labels_tag or not has_data_tag
                    if incomplete_regression_meta_data and incomplete_classification_meta_data:
                        raise TsFileParseException("a full set of metadata has not been provided before the data")

                    # Replace any missing values with the value specified
                    line = line.replace("?", replace_missing_vals_with)

                    # Check if we dealing with data that has timestamps
                    if timestamps:
                        # We're dealing with timestamps so cannot just split line on ':' as timestamps may contain one
                        has_another_value = False
                        has_another_dimension = False

                        timestamps_for_dimension = []
                        values_for_dimension = []

                        this_line_num_dimensions = 0
                        line_len = len(line)
                        char_num = 0

                        while char_num < line_len:
                            # Move through any spaces
                            while char_num < line_len and str.isspace(line[char_num]):
                                char_num += 1

                            # See if there is any more data to read in or if we should validate that read thus far

                            if char_num < line_len:

                                # See if we have an empty dimension (i.e. no values)
                                if line[char_num] == ":":
                                    if len(instance_list) < (this_line_num_dimensions + 1):
                                        instance_list.append([])

                                    instance_list[this_line_num_dimensions].append(pd.Series())
                                    this_line_num_dimensions += 1

                                    has_another_value = False
                                    has_another_dimension = True

                                    timestamps_for_dimension = []
                                    values_for_dimension = []

                                    char_num += 1
                                else:
                                    # Check if we have reached a class label
                                    if line[char_num] != "(" and target_labels:
                                        class_val = line[char_num:].strip()

                                        # if class_val not in class_val_list:
                                        #     raise TsFileParseException(
                                        #         "the class value '" + class_val + "' on line " + str(
                                        #             line_num + 1) + " is not valid")

                                        class_val_list.append(float(class_val))
                                        char_num = line_len

                                        has_another_value = False
                                        has_another_dimension = False

                                        timestamps_for_dimension = []
                                        values_for_dimension = []

                                    else:

                                        # Read in the data contained within the next tuple

                                        if line[char_num] != "(" and not target_labels:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " does not start with a '('")

                                        char_num += 1
                                        tuple_data = ""

                                        while char_num < line_len and line[char_num] != ")":
                                            tuple_data += line[char_num]
                                            char_num += 1

                                        if char_num >= line_len or line[char_num] != ")":
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " does not end with a ')'")

                                        # Read in any spaces immediately after the current tuple

                                        char_num += 1

                                        while char_num < line_len and str.isspace(line[char_num]):
                                            char_num += 1

                                        # Check if there is another value or dimension to process after this tuple

                                        if char_num >= line_len:
                                            has_another_value = False
                                            has_another_dimension = False

                                        elif line[char_num] == ",":
                                            has_another_value = True
                                            has_another_dimension = False

                                        elif line[char_num] == ":":
                                            has_another_value = False
                                            has_another_dimension = True

                                        char_num += 1

                                        # Get the numeric value for the tuple by reading from the end of the tuple data backwards to the last comma

                                        last_comma_index = tuple_data.rfind(',')

                                        if last_comma_index == -1:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains a tuple that has no comma inside of it")

                                        try:
                                            value = tuple_data[last_comma_index + 1:]
                                            value = float(value)

                                        except ValueError:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains a tuple that does not have a valid numeric value")

                                        # Check the type of timestamp that we have

                                        timestamp = tuple_data[0: last_comma_index]

                                        try:
                                            timestamp = int(timestamp)
                                            timestamp_is_int = True
                                            timestamp_is_timestamp = False
                                        except ValueError:
                                            timestamp_is_int = False

                                        if not timestamp_is_int:
                                            try:
                                                timestamp = float(timestamp)
                                                timestamp_is_float = True
                                                timestamp_is_timestamp = False
                                            except ValueError:
                                                timestamp_is_float = False

                                        if not timestamp_is_int and not timestamp_is_float:
                                            try:
                                                timestamp = timestamp.strip()
                                                timestamp_is_timestamp = True
                                            except ValueError:
                                                timestamp_is_timestamp = False

                                        # Make sure that the timestamps in the file (not just this dimension or case) are consistent

                                        if not timestamp_is_timestamp and not timestamp_is_int and not timestamp_is_float:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains a tuple that has an invalid timestamp '" + timestamp + "'")

                                        if previous_timestamp_was_float is not None and previous_timestamp_was_float and not timestamp_is_float:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                        if previous_timestamp_was_int is not None and previous_timestamp_was_int and not timestamp_is_int:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                        if previous_timestamp_was_timestamp is not None and previous_timestamp_was_timestamp and not timestamp_is_timestamp:
                                            raise TsFileParseException(
                                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                                    line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                        # Store the values

                                        timestamps_for_dimension += [timestamp]
                                        values_for_dimension += [value]

                                        #  If this was our first tuple then we store the type of timestamp we had

                                        if previous_timestamp_was_timestamp is None and timestamp_is_timestamp:
                                            previous_timestamp_was_timestamp = True
                                            previous_timestamp_was_int = False
                                            previous_timestamp_was_float = False

                                        if previous_timestamp_was_int is None and timestamp_is_int:
                                            previous_timestamp_was_timestamp = False
                                            previous_timestamp_was_int = True
                                            previous_timestamp_was_float = False

                                        if previous_timestamp_was_float is None and timestamp_is_float:
                                            previous_timestamp_was_timestamp = False
                                            previous_timestamp_was_int = False
                                            previous_timestamp_was_float = True

                                        # See if we should add the data for this dimension

                                        if not has_another_value:
                                            if len(instance_list) < (this_line_num_dimensions + 1):
                                                instance_list.append([])

                                            if timestamp_is_timestamp:
                                                timestamps_for_dimension = pd.DatetimeIndex(timestamps_for_dimension)

                                            instance_list[this_line_num_dimensions].append(
                                                pd.Series(index=timestamps_for_dimension, data=values_for_dimension))
                                            this_line_num_dimensions += 1

                                            timestamps_for_dimension = []
                                            values_for_dimension = []

                            elif has_another_value:
                                raise TsFileParseException(
                                    "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                        line_num + 1) + " ends with a ',' that is not followed by another tuple")

                            elif has_another_dimension and target_labels:
                                raise TsFileParseException(
                                    "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                        line_num + 1) + " ends with a ':' while it should list a class value")

                            elif has_another_dimension and not target_labels:
                                if len(instance_list) < (this_line_num_dimensions + 1):
                                    instance_list.append([])

                                instance_list[this_line_num_dimensions].append(pd.Series(dtype=np.float32))
                                this_line_num_dimensions += 1
                                num_dimensions = this_line_num_dimensions

                            # If this is the 1st line of data we have seen then note the dimensions

                            if not has_another_value and not has_another_dimension:
                                if num_dimensions is None:
                                    num_dimensions = this_line_num_dimensions

                                if num_dimensions != this_line_num_dimensions:
                                    raise TsFileParseException("line " + str(
                                        line_num + 1) + " does not have the same number of dimensions as the previous line of data")

                        # Check that we are not expecting some more data, and if not, store that processed above

                        if has_another_value:
                            raise TsFileParseException(
                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                    line_num + 1) + " ends with a ',' that is not followed by another tuple")

                        elif has_another_dimension and target_labels:
                            raise TsFileParseException(
                                "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                    line_num + 1) + " ends with a ':' while it should list a class value")

                        elif has_another_dimension and not target_labels:
                            if len(instance_list) < (this_line_num_dimensions + 1):
                                instance_list.append([])

                            instance_list[this_line_num_dimensions].append(pd.Series())
                            this_line_num_dimensions += 1
                            num_dimensions = this_line_num_dimensions

                        # If this is the 1st line of data we have seen then note the dimensions

                        if not has_another_value and num_dimensions != this_line_num_dimensions:
                            raise TsFileParseException("line " + str(
                                line_num + 1) + " does not have the same number of dimensions as the previous line of data")

                        # Check if we should have class values, and if so that they are contained in those listed in the metadata

                        if target_labels and len(class_val_list) == 0:
                            raise TsFileParseException("the cases have no associated class values")
                    else:
                        dimensions = line.split(":")
                        # If first row then note the number of dimensions (that must be the same for all cases)
                        if is_first_case:
                            num_dimensions = len(dimensions)

                            if target_labels:
                                num_dimensions -= 1

                            for dim in range(0, num_dimensions):
                                instance_list.append([])
                            is_first_case = False

                        # See how many dimensions that the case whose data in represented in this line has
                        this_line_num_dimensions = len(dimensions)

                        if target_labels:
                            this_line_num_dimensions -= 1

                        # All dimensions should be included for all series, even if they are empty
                        if this_line_num_dimensions != num_dimensions:
                            raise TsFileParseException("inconsistent number of dimensions. Expecting " + str(
                                num_dimensions) + " but have read " + str(this_line_num_dimensions))

                        # Process the data for each dimension
                        for dim in range(0, num_dimensions):
                            dimension = dimensions[dim].strip()

                            if dimension:
                                data_series = dimension.split(",")
                                data_series = [float(i) for i in data_series]
                                instance_list[dim].append(pd.Series(data_series))
                            else:
                                instance_list[dim].append(pd.Series())

                        if target_labels:
                            class_val_list.append(float(dimensions[num_dimensions].strip()))

            line_num += 1

    # Check that the file was not empty
    if line_num:
        # Check that the file contained both metadata and data
        complete_regression_meta_data = has_problem_name_tag and has_timestamps_tag and has_univariate_tag and has_target_labels_tag and has_data_tag
        complete_classification_meta_data = has_problem_name_tag and has_timestamps_tag and has_univariate_tag and has_class_labels_tag and has_data_tag

        if metadata_started and not complete_regression_meta_data and not complete_classification_meta_data:
            raise TsFileParseException("metadata incomplete")
        elif metadata_started and not data_started:
            raise TsFileParseException("file contained metadata but no data")
        elif metadata_started and data_started and len(instance_list) == 0:
            raise TsFileParseException("file contained metadata but no data")

        # Create a DataFrame from the data parsed above
        data = pd.DataFrame(dtype=np.float32)

        for dim in range(0, num_dimensions):
            data['dim_' + str(dim)] = instance_list[dim]

        # Check if we should return any associated class labels separately

        if target_labels:
            if return_separate_X_and_y:
                return data, np.asarray(class_val_list)
            else:
                data['class_vals'] = pd.Series(class_val_list)
                return data
        else:
            return data
    else:
        raise TsFileParseException("empty file")



def uniform_scaling(data, max_len):
    """
    This is a function to scale the time series uniformly
    :param data:
    :param max_len:
    :return:
    """
    seq_len = len(data)
    scaled_data = [data[int(j * seq_len / max_len)] for j in range(max_len)]

    return scaled_data


def to_fd(series, sample_points=None):
    """
    Convert a set of time series to functional data
    :param series: a set of time series with the shape of N x L
    :param sample_points: sample point of the series
    :return: functional data representation of the time series
    """
    from skfda import FDataGrid

    if sample_points is None:
        sample_points = [x for x in range(series.shape[1])]

    return FDataGrid(series, sample_points)



import logging
import setuptools
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError
from setuptools import Extension
import warnings

try:
    from Cython.Distutils import build_ext
except ImportError as e:
    warnings.warn(e.args[0])
    from setuptools.command.build_ext import build_ext

with open("README.rst", 'r') as f:
    long_description = f.read()

logging.basicConfig()
log = logging.getLogger(__file__)
ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError, IOError, SystemExit)


def get_version(fname):
    version = '0.0.1'
    package = 'llmabba'

    with open(fname) as f:
        for line in f:
            if line.startswith("__version__ = '"):
                version = line.split("'")[1]

            elif line.startswith("__name__ = '"):
                package = line.split("'")[1]

    return version, package
    # raise RuntimeError('Error in parsing version string.')


__version__, __package__ = get_version('llmabba/__init__.py')


class CustomBuildExtCommand(build_ext):
    """build_ext command for use when numpy headers are needed."""

    def run(self):
        import numpy
        self.include_dirs.append(numpy.get_include())
        build_ext.run(self)


setup_args = {'name': __package__,
              'packages': setuptools.find_packages(),
              'version': __version__,
              'cmdclass': {'build_ext': CustomBuildExtCommand},
              'setup_requires': ["cython", "numpy>=1.17.3"],
              'install_requires': ["numpy>=1.3.0",
                                   "scipy>=0.7.0",
                                   "requests",
                                   "pandas",
                                   "scikit-learn",
                                   "pychop",
                                   "cython>=0.27",
                                   "joblib>=1.1.1",
                                   "transformers",
                                   "peft",
                                   "trl",
                                   "datasets",
                                   "accelerate",
                                   "matplotlib"],

              'package_data': {__package__: [__package__]},
              'long_description': long_description,
              'author': "NA",
              'author_email': "xinyechenai@email.com",
              'classifiers': [
                  "Intended Audience :: Science/Research",
                  "Intended Audience :: Developers",
                  "Programming Language :: Python",
                  "Topic :: Software Development",
                  "Topic :: Scientific/Engineering",
                  "Operating System :: Microsoft :: Windows",
                  "Operating System :: Unix",
                  "Programming Language :: Python :: 3"
              ],
              'description': "LLM-ABBA: mining time series via symbolic approximation and large language models",
              'long_description_content_type': 'text/x-rst',
              'url': "https://github.com/inEXASCALE/llmabba",
              'license': 'BSD 3-Clause'
              }

comp = Extension('llmabba.comp',
                 sources=[__package__ + '/comp.pyx'])

agg = Extension('llmabba.agg',
                sources=[__package__  + '/agg.pyx'])

inverse = Extension('llmabba.inverse',
                    sources=[__package__  + '/inverse.pyx'])

try:
    setuptools.setup(
        **setup_args,
        ext_modules=[comp,
                     agg,
                     inverse
                     ],
    )

except ext_errors as ext_reason:
    log.warning(ext_reason)
    log.warning("The C extension could not be compiled.")
    if 'build_ext' in setup_args['cmdclass']:
        del setup_args['cmdclass']['build_ext']
    setuptools.setup(**setup_args)



__version__ = '0.0.2'
from .abba import XABBA, fastXABBA, fastXABBA_len, fastXABBA_inc

from .comp import compress
from .agg import aggregate 
from .inverse import *

from .llmabba import LLMABBA

__name__ =  'llmabba'
